#include "Parameters.h"
#include "Pedigree.h"
#include "math.h"
#include <iostream>
#include "FastFit.h"
#include "TraitTransformations.h"
#include "OutputKin.h"
#include "TransformResiduals.h"
#include "PEDtoVCF.h"
#include "VT.h"
#include "Burden.h"
#include "SingleVariantTest.h"
#include <stdio.h>
#include <time.h>
#include "Error.h"
#include "AutoFit.h"
#include "GroupFromAnnotation.h"
#define MATHLIB_STANDALONE
#include "Rmath.h"
#include "DataQC.h"
#include "BgzfFileType.h"

int main(int argc, char ** argv)
{
   printf("\nfamrvtest %s -- An associaiton tool for family and unrealted samples.\n"
	 "          (c) 2012-2014 Shuang Feng, Dajiang Liu, Goncalo Abecasis\n\n", VERSION);

   PhoneHome::allThinning = 100;
   bool noeof = false;
   time_t initialTime = time(0);
   String pedfile, datfile;
   double tol = 0.0001;

   BEGIN_LONG_PARAMETERS(additional)
      LONG_PARAMETER_GROUP("Kinship Options")
      LONG_PARAMETER("kinGeno", &FastTransform::empKin)
      LONG_PARAMETER("kinPedigree", &FastTransform::pedKin)
      LONG_STRINGPARAMETER("kinFile", &FastTransform::readInEmp)
      LONG_PARAMETER("kinSave", &OutputKin::outputKin)
      LONG_DOUBLEPARAMETER("kinMaf", &KinshipEmp::q)
      LONG_DOUBLEPARAMETER("kinMiss", &KinshipEmp::miss)
      //LONG_PARAMETER("outputX", &OutputKin::outputKinX)
      //LONG_PARAMETER_GROUP("Pre-processing Options")
      //LONG_PARAMETER("preProcess", &VT::preProcess)
      //LONG_INTPARAMETER("chopSize", &VT::chop)
      //LONG_STRINGPARAMETER("saveChopTo", &VT::chopTo)
      LONG_PARAMETER_GROUP("Input/Output Options")
      LONG_STRINGPARAMETER("ped", &pedfile)
      LONG_STRINGPARAMETER("dat", &datfile)
      LONG_STRINGPARAMETER("vcf", &GroupFromAnnotation::vcfInput)
      LONG_PARAMETER("noeof", &noeof)
      LONG_STRINGPARAMETER("groupFile", &GroupFromAnnotation::groupFile)
      LONG_PARAMETER("longOutput", &Burden::fullResult)
      LONG_STRINGPARAMETER("prefix", &Burden::outputFile)
      LONG_PARAMETER_GROUP("VC Options")
      LONG_PARAMETER("inverseNormal", &FastFit::inverseNormal)
      LONG_PARAMETER("useCovariates", &FastFit::useCovariates)
      LONG_STRINGPARAMETER("traitName", &FastFit::traitName)
      LONG_PARAMETER_GROUP("Single Variant Tests")
      LONG_PARAMETER("SingleVarScore", &SVT::SCORE)
      LONG_PARAMETER_GROUP("Burden Tests")
      //LONG_PARAMETER("Burden", &Burden::burden)
      LONG_PARAMETER("burden", &Burden::CMC_counts)
      LONG_PARAMETER("MB", &Burden::MB)
      LONG_PARAMETER("VT", &VT::asymptotic)
      LONG_PARAMETER("SKAT_BETA", &Burden::SKAT_BETA)
      LONG_PARAMETER("SKAT_UW", &Burden::SKAT_UW)
      LONG_PARAMETER("SKAT_MB", &Burden::SKAT_MB)
      //LONG_PARAMETER("LRT", &Burden::LRT)
      //LONG_PARAMETER("Wald", &Burden::WALD)
      LONG_PARAMETER_GROUP("Other Options")
      LONG_STRINGPARAMETER("freqFile", &GroupFromAnnotation::freqFile)
      LONG_STRINGPARAMETER("annotation", &GroupFromAnnotation::function)
      //LONG_STRINGPARAMETER("chrom", &Burden::chromosome)
      //LONG_DOUBLEPARAMETER("mafMin", &Burden::mafMin)
      LONG_DOUBLEPARAMETER("maf", &Burden::mafMax)
      LONG_DOUBLEPARAMETER("miss", &Burden::miss)
      LONG_INTPARAMETER("simulate", &SVT::simulate)
      LONG_PARAMETER("dosage", &SVT::dosage)
      LONG_PARAMETER("founderFreq", &SVT::FounderFreq)
      LONG_PHONEHOME(VERSION)
      //LONG_PARAMETER("h2Only", &FastFit::h2Only)
      BEGIN_LEGACY_PARAMETERS()
      LONG_STRINGPARAMETER("xLabel", &SVT::xLabel)
      LONG_INTPARAMETER("Xstart", &SVT::Xstart)
      LONG_INTPARAMETER("Xend", &SVT::Xend)
      LONG_PARAMETER("fitX", &AutoFit::fitX)
      LONG_PARAMETER("noStop", &VT::noStep)
      LONG_INTPARAMETER("maleLabel", &SVT::maleLabel)
      LONG_INTPARAMETER("femaleLabel", &SVT::femaleLabel)
      LONG_INTPARAMETER("permuteMin", &VT::permuteMin)
      LONG_INTPARAMETER("permuteMax", &VT::permuteMax)
      LONG_PARAMETER("VTpermute", &VT::permuteBool)
      LONG_PARAMETER("VTMB", &VT::VTwtMB)
      LONG_PARAMETER("VTCMC_binary", &VT::VTwtCMC_binary)
      LONG_PARAMETER("VTCMC_counts", &VT::VTwtCMC_counts)
      LONG_PARAMETER("CMC_binary", &Burden::CMC_binary)
      LONG_PARAMETER("LinScore", &Burden::LinScoreAssoc)
      LONG_PARAMETER("SingleVarLRT", &SVT::LRT)
      LONG_PARAMETER("SingleVarWald", &SVT::WALD)
      LONG_PARAMETER("fitSharedEnv", &AutoFit::fitSharedEnvironment)
      END_LONG_PARAMETERS();

   ParameterList pl;
   pl.Add(new LongParameters("Options:", additional));

   pl.Read(argc, argv);

   PhoneHome::checkVersion("raremetalworker",VERSION);

   pl.Status();
   OutputKin::filePrefix = Burden::outputFile; 

   /*
      if(!SVT::test && !Burden::MB && !Burden::CMC_counts && !Burden::CMC_binary)
      {
      error("ERROR! Rare variant test names need to be entered.\nOptions are: --svTest, --CMC_binary, --CMC_counts, and --MB.\n\n");
      }
    */

   if(noeof)
   {
      // Set that the eof block is not required.
      BgzfFileType::setRequireEofBlock(false);
   }

   Pedigree ped;
   if(SVT::simulate>0)
   {
      printf("WARNING: %d permutations are requested for single variant tests. This might take long.\n",SVT::simulate);
      String label;
      label = "dummy";
      int markerId = ped.GetMarkerID(label);
   }

   ped.Prepare(datfile);
   ped.Load(pedfile);

   if(ped.markerCount>0)
      SVT::genoFromPed = true;
   else if(!SVT::dosage) 
      SVT::genoFromVCF=true;

   if(Burden::burden || Burden::SKAT_BETA || Burden::SKAT_UW || Burden::SKAT_MB || VT::asymptotic || VT::permuteBool)
   {
      if(GroupFromAnnotation::vcfInput =="" && GroupFromAnnotation::groupFile == "")
	 error("FATAL ERROR: annotated VCF file or group file needed for Burden tests.\n");
   }
   FILE * log;
   String logFile;
   if(Burden::outputFile != "")
   {
      if(Burden::outputFile.FastFindLastChar('.')!=-1 || Burden::outputFile.FastFindLastChar('/')!=-1)
	 logFile = Burden::outputFile + "famRvTest.log";
      else
	 logFile = Burden::outputFile + ".famRvTest.log";
   }
   else
      logFile = "famRvTest.log";

   log = freopen (logFile,"wt",stderr);

   time_t now;
   time(&now);
   printf("Analysis started at: %s\n", ctime(&now));
   fprintf(log,"Analysis started at: %s\n", ctime(&now));

   try
   {
      SanityCheck checkData;
      checkData.Check(ped,log);

      AutoFit fitEngine;
      FastFit engine;
      FastTransform trans;
      KinshipEmp kin_emp;
      //check if the sample IDs from PED and VCF files are consistent
      trans.ScreenSampleID(ped,FastFit::useCovariates);

      //setup empirical kinship if applicable
      if(FastTransform::empKin)
	 kin_emp.SetupEmpKin(ped,trans.genotypedSamplePED,trans.genotypedSampleVCF,trans.samplePEDIDHash,checkData.skippedSNPs,checkData.chromosomeVCF,log);

      StringArray traits;
      if(!FastFit::traitName.IsEmpty())
      {
	 if(FastFit::traitName == "traitName.txt")
	 {
	    FILE * trFile = fopen(FastFit::traitName,"r");
	    String buffer;
	    while (!feof(trFile))
	    {
	       buffer.ReadLine(trFile);
	       traits.AddTokens(buffer, SEPARATORS);
	    }
	    fclose(trFile);
	 }
	 else
	 {
	    traits.AddTokens(FastFit::traitName,"/");
	 }
      }
      else 
      {
	 traits = ped.traitNames;
      }

      for(int i=0;i<traits.Length();i++)
      {
	 FastFit::traitName.Copy(traits[i]);
	 IFILE dummy;
	 dummy = NULL;
	 engine.FitModels(dummy,dummy,dummy,ped,tol,trans,kin_emp,log);
	 /*
	    if(FastFit::h2Only)
	    {
	    FILE * h2out;
	    String file = Burden::outputFile + "_" + FastFit::traitName + "." + "heritability";
	    h2out=fopen(file,"wt");
	    double h2=engine.sigma_g2Hat/trans.traitVar;
	    printf("Heritability for trait %s saved in %s.\n",FastFit::traitName.c_str(),file.c_str());
	    fprintf(h2out,"%s\t%g\n",FastFit::traitName.c_str(),h2);
	    fclose(h2out);
	    continue;
	    }
	  */
	 //Now do the burden tests

	 SVT svt;
	 svt.Run(ped,trans,engine,traits[i],log);

	 GroupFromAnnotation group;
	 group.Run();

	 if(Burden::SKAT_BETA || Burden::SKAT_UW || Burden::SKAT_MB || Burden::burden)
	 {
	    Burden burden;
	    burden.Run(ped,trans,engine,traits[i],group,log);	
	 }

	 VT vt;
	 vt.Run(ped,trans,engine,traits[i],group,log);
      }

      time_t endTime = time(0);
      int timeSec = (int) (endTime - initialTime);
      double timeHour = timeSec/3600.0;
      printf("Analysis took %d seconds (~ %.1f hours).\n",timeSec, timeHour);
      fprintf(log,"Analysis took %d seconds (~ %.1f hours).\n",timeSec, timeHour);

      fclose(log);
   }
   catch(std::exception& e)
   {
      printf("Exiting, exception thrown: %s\n\n",e.what());
      fprintf(log,"Exiting, exception thrown: %s\n\n",e.what());
      PhoneHome::completionStatus("Exception");
      return(-1);
   }
   PhoneHome::completionStatus("SUCCESS");
   return 0;
}
